/* Modified by Ian Miller and Waseh Ahmad
 * File:   main.c
 * Author: Matthew Watkins & Syed Tahmid Mahbub
 *
 * Created on October 19, 2014
 */


//IMPORTANT: Include config.h, then protothreads, then everything else
#include "config.h"
#include "pt_cornell_1_2.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "plib.h"
#include "xc.h"
#include "adc_intf.h"
#include "tft_master.h"
#include "tft_gfx.h"

#define SYS_FREQ 40000000

#define SERVO_PERIOD_US 20000
#define SERVO_PERIOD_CENTER 1500
#define SERVO_TIME_TO_REG(us) (((SYS_FREQ/100000)/16)*((us)/10))

#define MOTOR_SPEED 300

#define BUTTON_1L   BIT_1
#define BUTTON_1R   BIT_2
#define BUTTON_2L   BIT_3
#define BUTTON_2R   BIT_4


#define IP_GATEWAY "192.168.173.1"
#define NETMASK "255.255.255.0"
#define SERVER_IP "192.168.173.184"
#define CONTROLLER_IP "192.168.173.75"


volatile char buffer[64];

//Keep track of edge
volatile uint32_t endstop_1[2];
volatile uint32_t endstop_2[2];

//Variable to keep track of counts
volatile uint32_t location[2];
volatile uint32_t target[2];
//Variable to keep track of the current direction of rotation
typedef enum {
  CW, CCW
} DIRECTION;
volatile DIRECTION direction[2];

//Select the current motor
volatile uint8_t curMotor;

//State variable for overall system
volatile enum {
    NORMAL, CALIB_CW, CALIB_CCW, GOTO_TARGET, OPEN_1, OPEN_2, CLOSE_1, CLOSE_2
} mode;

//Protothreads
static struct pt pt_cmd_line, pt_buttons, pt_io,pt_start_esp,pt_receive_esp;

void selectMotor(uint8_t m);

//Interface functions
void increaseBrightness();
void decreaseBrightness();
void stopMotion();
void stopOverride();

//Handle motion when there is a set target for the current motor
void moveToTarget() {
    if (location[curMotor] < target[curMotor]) {
        direction[curMotor] = CW;
        SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER+MOTOR_SPEED));
        ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
        ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
    } else if (location[curMotor] > target[curMotor]) {
        direction[curMotor] = CCW;
        SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER-MOTOR_SPEED));
        ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
        ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
    } else {
        //We have reached the end of travel
        if (mode == OPEN_1) {
            //Switch to next motor
            delay_ms(100);
            selectMotor(1);
            target[curMotor] = (uint32_t)(endstop_2[curMotor]+3);
            mode = OPEN_2;
            WriteTimer3(0);
            moveToTarget();
        } else if (mode == CLOSE_1) {
            //Switch to next motor
            delay_ms(100);
            selectMotor(0);
            target[curMotor] = (uint32_t)(endstop_1[curMotor]-3);
            mode = CLOSE_2;
            WriteTimer3(0);
            moveToTarget();
        } else {
            //We are done, shutdown motors
            mode = NORMAL;
            SetDCOC1PWM(0);
            //Turn off interrupts
            ConfigIntCapture1(IC_INT_OFF);
            ConfigIntTimer3(T3_INT_OFF);
            stopOverride();
        }
    }
}

//ISR for input capture: called when a "tick" received
void __ISR(_INPUT_CAPTURE_1_VECTOR,ipl2auto) IC1Handler(void){
    //We have to read the capture value to clear the ISR
    volatile uint32_t val = mIC1ReadCapture();

    //Update position
    if (direction[curMotor] == CW) {
        location[curMotor]++;
    } else {
        location[curMotor]--;
    }
    
    //Target search
    if (mode == GOTO_TARGET || 
            mode == OPEN_1 || mode == OPEN_2 ||
            mode == CLOSE_1 || mode == CLOSE_2) {
        moveToTarget();
    }
    //Reset timer to indicate we are still turning
    WriteTimer3(0);
    mIC1ClearIntFlag();
}

//ISR for triggering when we haven't had a Hall Effect trigger in a while
//Means we have either jammed or reached the end of travel
void __ISR(_TIMER_3_VECTOR,ipl2auto) T3Handler(void) {
    //Handle calibration behavior
    if (mode == CALIB_CW) {
        //Switch direction to find other endstop
        endstop_1[curMotor] = location[curMotor];
        direction[curMotor] = CCW;
        mode = CALIB_CCW;
        SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER-MOTOR_SPEED));
    } else if (mode == CALIB_CCW) {
        //We are done, halt motion
        endstop_2[curMotor] = location[curMotor];
        direction[curMotor] = CW;
        mode = NORMAL;
        ConfigIntCapture1(IC_INT_OFF);
        ConfigIntTimer3(T3_INT_OFF);
        SetDCOC1PWM(0);
    } else {
        //Stop motors (jam detected)
        mode = NORMAL;
        ConfigIntCapture1(IC_INT_OFF);
        ConfigIntTimer3(T3_INT_OFF);
        SetDCOC1PWM(0);
        stopOverride();
    }
    
    mT3ClearIntFlag();
}

//Interface methods for network to call
void increaseBrightness() {
    mPORTBToggleBits(BIT_7);
    selectMotor(0);
    target[curMotor] = (uint32_t)(endstop_2[curMotor]+3);
    mode = OPEN_1;
    //Directly call ISR to start motion
    WriteTimer3(0);
    moveToTarget();
    WriteTimer3(0);
}

void decreaseBrightness() {
    mPORTBToggleBits(BIT_7);
    selectMotor(1);
    target[curMotor] = (uint32_t)(endstop_1[curMotor]-3);
    mode = CLOSE_1;
    //Directly call ISR to start motion
    WriteTimer3(0);
    moveToTarget();
    WriteTimer3(0);
}

void stopMotion() {
    mode = NORMAL;
    ConfigIntCapture1(IC_INT_OFF);
    ConfigIntTimer3(T3_INT_OFF);
    SetDCOC1PWM(0);
}

//Transmit signal to tell controller that we are at the end of travel
void stopOverride() {
    
    esp8266_untarnished_send("S!!");
    //printf("Halted\n\r");
}

//State variables for button state machine system
volatile uint32_t lastButtonPressed = 0;
volatile uint32_t timeSincePressed = 0;

//Logic to handle a button press
void handlePress(uint32_t button, uint8_t motor, DIRECTION dir) {
    //Double press?
    selectMotor(motor);
    if (timeSincePressed < 5 && timeSincePressed > 0 && lastButtonPressed == button) {
        if (dir == CW) {
            target[curMotor] = (uint32_t)(endstop_1[curMotor]-3);
        } else {
            target[curMotor] = (uint32_t)(endstop_2[curMotor]+3);
        }
        mode = GOTO_TARGET;
        //Directly call ISR to start motion
        moveToTarget();
        //mPORTBToggleBits(BIT_9);
        WriteTimer3(0);
    } else {
        direction[curMotor] = dir;
        if (location[curMotor] < (endstop_1[curMotor]-3) && dir == CW) {
            SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER+MOTOR_SPEED));
            ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
            ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
        } else if (location[curMotor] > (endstop_2[curMotor]+3) && dir == CCW) {
            SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER-MOTOR_SPEED));
            ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
            ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
        } else {
            //Stop if we hit the edge
            ConfigIntCapture1(IC_INT_OFF);
            ConfigIntTimer3(T3_INT_OFF);
            SetDCOC1PWM(0);
        }
    }
    lastButtonPressed = button;
    timeSincePressed = 0;
}

static PT_THREAD (protothread_buttons(struct pt *pt)){
    PT_BEGIN(pt);
    while (1) {
        //Read in button states
        uint32_t button_1l = !PORTReadBits(IOPORT_A, BUTTON_1L);
        uint32_t button_1r = !PORTReadBits(IOPORT_A, BUTTON_1R);
        uint32_t button_2l = !PORTReadBits(IOPORT_A, BUTTON_2L);
        uint32_t button_2r = !PORTReadBits(IOPORT_A, BUTTON_2R);
        
        if (button_1l && button_1r) {
            //Calibrate first DOF
            selectMotor(0);
            SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER+MOTOR_SPEED));
            WriteTimer3(0);
            //Turn on interrupts
            ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
            ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
            direction[curMotor] = CW;
            mode = CALIB_CW;
        } else if (button_2l && button_2r) {
            //Calibrate second DOF
            selectMotor(1);
            SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER+MOTOR_SPEED));
            WriteTimer3(0);
            //Turn on interrupts
            ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
            ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
            direction[curMotor] = CW;
            mode = CALIB_CW;
        } else if (mode == NORMAL) { //Are we not calibrating? 
            if (button_1l) {
                mPORTBToggleBits(BIT_7);
                handlePress(BUTTON_1L, 0, CW);
            } else if (button_2l) {
                mPORTBToggleBits(BIT_7);
                handlePress(BUTTON_2L, 1, CW);
            } else if (button_1r) {
                mPORTBToggleBits(BIT_7);
                handlePress(BUTTON_1R, 0, CCW);
            } else if (button_2r) {
                mPORTBToggleBits(BIT_7);
                handlePress(BUTTON_2R, 1, CCW);
            } else {
                if (timeSincePressed < 20){
                    timeSincePressed++;
                }
                //Stop motor if just released
                if (timeSincePressed == 1) {
                    ConfigIntCapture1(IC_INT_OFF);
                    ConfigIntTimer3(T3_INT_OFF);
                    SetDCOC1PWM(0);
                    stopOverride();
                }
            }
        }
        //Waiting a fair amount of time b/w samples achieves basic debouncing
        PT_YIELD_TIME_msec(100);
    }
    PT_END(pt);
}

static uint32_t recv_val = 0;
static PT_THREAD (protothread_cmd_line(struct pt *pt)){
    PT_BEGIN(pt);
    while(1){
		//Print prompt via UART
		strncpy(PT_send_buffer, "-> ", sizeof(PT_send_buffer));
		PT_SPAWN(pt, &pt_io, PutSerialBuffer(&pt_io) );
		//Get terminal input
		//PT_SPAWN will wait until the thread returns, loading input into PT_term_buffer
		PT_SPAWN(pt, &pt_io, PT_GetSerialBuffer(&pt_io) );
		
		//Make sure command is long enough, then try to parse
		if (num_char >= 3) {
			sscanf(&PT_term_buffer[2], "%u", &recv_val);
		} else {
            strncpy(PT_send_buffer, "Invalid cmd\n\r", sizeof(PT_send_buffer));
		}
		
		switch (PT_term_buffer[0]) {
            case 'c': //Calibration command
                selectMotor(recv_val);
                SetDCOC1PWM(SERVO_TIME_TO_REG(SERVO_PERIOD_CENTER+MOTOR_SPEED));
                ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
                ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
                WriteTimer3(0);
                direction[curMotor] = CW;
                mode = CALIB_CW;
                break;
            case 'p': //Print endstop locations
                sprintf(buffer,"0: %u, %u 1: %u, %u\n\r",
                        endstop_1[0]-endstop_2[0], location[0]-endstop_2[0],
                        endstop_1[1]-endstop_2[1], location[1]-endstop_2[1]);
                strncpy(PT_send_buffer, buffer, sizeof(PT_send_buffer));
                PT_SPAWN(pt, &pt_io, PutSerialBuffer(&pt_io) );
                break;
            case 's':
                selectMotor(recv_val);
                break;
            case 'g':
                target[curMotor] = (uint32_t)(endstop_2[curMotor]+recv_val);
                mode = GOTO_TARGET;
                //Directly call ISR to start motion
                moveToTarget();
                WriteTimer3(0);
                break;
            case 'i':
                increaseBrightness();
                break;
            case 'd':
                decreaseBrightness();
                break;
            case 'h':
                stopMotion();
                break;
        }
    }
    
    PT_END(pt);
}

//Select which motor we are controlling
void selectMotor(uint8_t m) {
    curMotor = m;
    INTDisableInterrupts();
    if (m==0) {
       PPSOutput(1,RPB3,NULL);
       PPSOutput(1,RPA0,OC1);
       PPSInput(3,IC1,RPB13);
    } else {
       PPSOutput(1,RPB3,OC1);
       PPSInput(3,IC1,RPB2);
       PPSOutput(1,RPA0,NULL);
    }
    INTEnableSystemMultiVectoredInt();
}

//================================================================================
////protothread to receive commands from the web interface
static char receive_buffer[60];
static uint8_t recv_data;
//static uint16_t receive_size;
static PT_THREAD (protothread_receive_esp(struct pt *pt)){
    PT_BEGIN(pt);
    
    while(1){
        //Hang out until we receive data from the ESP
        PT_YIELD_UNTIL(pt, UARTReceivedDataIsAvailable(UART2));//yield until data is available to be received
                
        //esp8266_receive_all(receive_buffer,20/*receive_size*/);
        recv_data = UARTGetDataByte(UART2);   
           
        //For debugging: toggle LED when command received
        mPORTBToggleBits(BIT_9);

        //Take action depending on command
        if(recv_data == 'I'){
            increaseBrightness();
        }else if(recv_data == 'D'){
            decreaseBrightness();
        }else{
            stopMotion();
        }
        //check for command from the MOtor controller
    }
    
    PT_END(pt);
}



// <editor-fold defaultstate="collapsed" desc="Start ESP Thread which runs once">
//===============================================================================
//protothread to start the esp
static PT_THREAD (protothread_start_esp(struct pt *pt)){
    PT_BEGIN(pt);
    
    tft_setCursor(10, 10); 
    sprintf(buffer,"Restarting ESP");
    tft_fillRect(10, 10, 400, 100, ILI9341_BLACK);
    tft_writeString(buffer);
    
    Initialize_ESP8266();
    esp8266_restart();
    delay_ms(2000);
    do{
        delay_ms(1500);    
    }while(!esp8266_isStarted());
    
    delay_ms(1500); 
    
    //place esp8266 in station mode
    esp8266_mode(1);
    delay_ms(1500);
    
    //setup the ip address of the CIP
    esp8266_change_ip(CONTROLLER_IP,IP_GATEWAY,NETMASK);
    //connect to network
    esp8266_connect("waseh","12345678");
    
    tft_setCursor(10, 10);

    delay_ms(1500);
    
    //connect to server
    esp8266_start(1,SERVER_IP,80);
    delay_ms(1500);
    
    //enable untarnished mode
    esp8266_untarnished_mode();
    delay_ms(1500);
    
    //go to untarnished mode for sending
    esp8266_untarnished_send_start();
    delay_ms(1500);
    
    PT_END(pt);
}
// </editor-fold>

//================================================================================

int main(int argc, char** argv) {
    
    SYSTEMConfigPerformance(SYS_FREQ);

    CM1CON = 0; CM2CON = 0; ANSELA = 0; ANSELB = 0;
    
    configureADC();

    PT_setup();
    PT_INIT(&pt_cmd_line);
    PT_INIT(&pt_buttons);
    PT_INIT(&pt_start_esp);
    PT_INIT(&pt_receive_esp);

    //setup output compare to generate PWM
    OpenOC1(OC_ON | OC_TIMER_MODE16 | OC_TIMER2_SRC | OC_PWM_FAULT_PIN_DISABLE,0,0);
    //use RPA0 for output compare out
    PPSOutput(1,RPA0,OC1);
    //Timer 2 used for OC        
    OpenTimer2(T2_ON | T2_PS_1_16, SERVO_TIME_TO_REG(SERVO_PERIOD_US));
    SetDCOC1PWM(0);
    
    //Input Capture for Hall Effect
    //Maximum prescalar.  We aren't all that interested in super-precision, and only get a click ~10Hz
    //With a 1:256 prescalar and 16 bit timer, we can do a min of ~0.5Hz
    OpenTimer3(T3_ON | T3_PS_1_256, 0xFFFF);
    OpenCapture1(IC_ON | IC_CAP_16BIT | IC_TIMER3_SRC |  IC_EVERY_EDGE | IC_INT_1CAPTURE);
    ConfigIntCapture1(IC_INT_ON | IC_INT_PRIOR_2);
    //PPS for Hall Effect input
    PPSInput(3,IC1,RPB13);
    //Interrupt for overflow detection
    ConfigIntTimer3(T3_INT_ON | T3_INT_PRIOR_2);
    
    //No motion
    SetDCOC1PWM(SERVO_TIME_TO_REG(0));
    ConfigIntCapture1(IC_INT_OFF);
    ConfigIntTimer3(T3_INT_OFF);
    
    //Setup input buttons
    //set up internal resistors (pullups)
    mCNAOpen(CNA_OFF,CNA1_ENABLE|CNA2_ENABLE|CNA3_ENABLE|CNA4_ENABLE, 
            CNA1_PULLUP_ENABLE|CNA2_PULLUP_ENABLE|CNA3_PULLUP_ENABLE|CNA4_PULLUP_ENABLE);
    
    //set initial switch 
    mPORTASetPinsDigitalIn(BUTTON_1L | BUTTON_1R | BUTTON_2L | BUTTON_2R);
    
    //For debugging
    mPORTBSetPinsDigitalOut(BIT_7 | BIT_9);
    
    selectMotor(0);
    
    //Initialize global variables
    location[0] = (0xFFFFFFFF/2);
    endstop_1[0] = 0xFFFFFFFF;
    endstop_2[0] = 0;
    direction[0] = CW;
    location[1] = (0xFFFFFFFF/2);
    endstop_1[1] = 0xFFFFFFFF;
    endstop_2[1] = 0;
    direction[1] = CW;
    
    mode = NORMAL;
    
    //need to run this one to start esp
    PT_SCHEDULE(protothread_start_esp(&pt_start_esp));

    //=========================================
    //Global Interrupt Enable
    INTEnableSystemMultiVectoredInt();

    //Let's thread!
    while (1){
        PT_SCHEDULE(protothread_buttons(&pt_buttons));
        PT_SCHEDULE(protothread_receive_esp(&pt_receive_esp));
        //PT_SCHEDULE(protothread_cmd_line(&pt_cmd_line));
    }
}
